@@ -14,7 +14,7 @@ try:
 except ImportError:
     from urllib.parse import urlparse
 
-from traitlets import Unicode, Integer, Bool, default, observe
+from traitlets import Unicode, Integer, default, observe
 
 from jupyter_core.application import JupyterApp, base_aliases
 from jupyter_client.kernelspec import KernelSpecManager
@@ -35,8 +35,8 @@ from .services.activity.manager import ActivityManager
 from .services.kernels.manager import SeedingMappingKernelManager
 
 # Only present for generating help documentation
-from kernel_gateway.notebook_http import NotebookHTTPPersonality
-from kernel_gateway.jupyter_websocket import JupyterWebsocketPersonality
+from .notebook_http import NotebookHTTPPersonality
+from .jupyter_websocket import JupyterWebsocketPersonality
 
 # Add additional command line aliases
 aliases = dict(base_aliases)
@@ -45,9 +45,13 @@ aliases.update({
     'port': 'KernelGatewayApp.port',
     'port_retries': 'KernelGatewayApp.port_retries',
     'api': 'KernelGatewayApp.api',
-    'seed_uri': 'KernelGatewayApp.seed_uri'
+    'seed_uri': 'KernelGatewayApp.seed_uri',
+    'keyfile': 'KernelGatewayApp.keyfile',
+    'certfile': 'KernelGatewayApp.certfile',
+    'client-ca': 'KernelGatewayApp.client_ca'
 })
 
+
 class KernelGatewayApp(JupyterApp):
     """Application that provisions Jupyter kernels and proxies HTTP/Websocket
     traffic to the kernels.
@@ -58,7 +62,7 @@ class KernelGatewayApp(JupyterApp):
     - starts the Tornado event loop
     """
     name = 'jupyter-kernel-gateway'
-    version =  __version__
+    version = __version__
     description = """
         Jupyter Kernel Gateway
 
@@ -77,6 +81,7 @@ class KernelGatewayApp(JupyterApp):
     port = Integer(port_default_value, config=True,
         help="Port on which to listen (KG_PORT env var)"
     )
+
     @default('port')
     def port_default(self):
         return int(os.getenv(self.port_env, self.port_default_value))
@@ -86,6 +91,7 @@ class KernelGatewayApp(JupyterApp):
     port_retries = Integer(port_retries_default_value, config=True,
         help="Number of ports to try if the specified port is not available (KG_PORT_RETRIES env var)"
     )
+
     @default('port_retries')
     def port_retries_default(self):
         return int(os.getenv(self.port_retries_env, self.port_retries_default_value))
@@ -95,6 +101,7 @@ class KernelGatewayApp(JupyterApp):
     ip = Unicode(ip_default_value, config=True,
         help="IP address on which to listen (KG_IP env var)"
     )
+
     @default('ip')
     def ip_default(self):
         return os.getenv(self.ip_env, self.ip_default_value)
@@ -104,6 +111,7 @@ class KernelGatewayApp(JupyterApp):
     base_url_default_value = '/'
     base_url = Unicode(base_url_default_value, config=True,
         help="""The base path for mounting all API resources (KG_BASE_URL env var)""")
+
     @default('base_url')
     def base_url_default(self):
         return os.getenv(self.base_url_env, self.base_url_default_value)
@@ -113,6 +121,7 @@ class KernelGatewayApp(JupyterApp):
     auth_token = Unicode(config=True,
         help='Authorization token required for all requests (KG_AUTH_TOKEN env var)'
     )
+
     @default('auth_token')
     def _auth_token_default(self):
         return os.getenv(self.auth_token_env, '')
@@ -122,6 +131,7 @@ class KernelGatewayApp(JupyterApp):
     allow_credentials = Unicode(config=True,
         help='Sets the Access-Control-Allow-Credentials header. (KG_ALLOW_CREDENTIALS env var)'
     )
+
     @default('allow_credentials')
     def allow_credentials_default(self):
         return os.getenv(self.allow_credentials_env, '')
@@ -130,6 +140,7 @@ class KernelGatewayApp(JupyterApp):
     allow_headers = Unicode(config=True,
         help='Sets the Access-Control-Allow-Headers header. (KG_ALLOW_HEADERS env var)'
     )
+
     @default('allow_headers')
     def allow_headers_default(self):
         return os.getenv(self.allow_headers_env, '')
@@ -138,6 +149,7 @@ class KernelGatewayApp(JupyterApp):
     allow_methods = Unicode(config=True,
         help='Sets the Access-Control-Allow-Methods header. (KG_ALLOW_METHODS env var)'
     )
+
     @default('allow_methods')
     def allow_methods_default(self):
         return os.getenv(self.allow_methods_env, '')
@@ -146,6 +158,7 @@ class KernelGatewayApp(JupyterApp):
     allow_origin = Unicode(config=True,
         help='Sets the Access-Control-Allow-Origin header. (KG_ALLOW_ORIGIN env var)'
     )
+
     @default('allow_origin')
     def allow_origin_default(self):
         return os.getenv(self.allow_origin_env, '')
@@ -154,6 +167,7 @@ class KernelGatewayApp(JupyterApp):
     expose_headers = Unicode(config=True,
         help='Sets the Access-Control-Expose-Headers header. (KG_EXPOSE_HEADERS env var)'
     )
+
     @default('expose_headers')
     def expose_headers_default(self):
         return os.getenv(self.expose_headers_env, '')
@@ -162,6 +176,7 @@ class KernelGatewayApp(JupyterApp):
     max_age = Unicode(config=True,
         help='Sets the Access-Control-Max-Age header. (KG_MAX_AGE env var)'
     )
+
     @default('max_age')
     def max_age_default(self):
         return os.getenv(self.max_age_env, '')
@@ -171,6 +186,7 @@ class KernelGatewayApp(JupyterApp):
         allow_none=True,
         help='Limits the number of kernel instances allowed to run by this gateway. Unbounded by default. (KG_MAX_KERNELS env var)'
     )
+
     @default('max_kernels')
     def max_kernels_default(self):
         val = os.getenv(self.max_kernels_env)
@@ -181,6 +197,7 @@ class KernelGatewayApp(JupyterApp):
         allow_none=True,
         help='Runs the notebook (.ipynb) at the given URI on every kernel launched. No seed by default. (KG_SEED_URI env var)'
     )
+
     @default('seed_uri')
     def seed_uri_default(self):
         return os.getenv(self.seed_uri_env)
@@ -190,6 +207,7 @@ class KernelGatewayApp(JupyterApp):
         allow_none=True,
         help='Number of kernels to prespawn using the default language. No prespawn by default. (KG_PRESPAWN_COUNT env var)'
     )
+
     @default('prespawn_count')
     def prespawn_count_default(self):
         val = os.getenv(self.prespawn_count_env)
@@ -198,6 +216,7 @@ class KernelGatewayApp(JupyterApp):
     default_kernel_name_env = 'KG_DEFAULT_KERNEL_NAME'
     default_kernel_name = Unicode(config=True,
         help='Default kernel name when spawning a kernel (KG_DEFAULT_KERNEL_NAME env var)')
+
     @default('default_kernel_name')
     def default_kernel_name_default(self):
         # defaults to Jupyter's default kernel name on empty string
@@ -206,6 +225,7 @@ class KernelGatewayApp(JupyterApp):
     force_kernel_name_env = 'KG_FORCE_KERNEL_NAME'
     force_kernel_name = Unicode(config=True,
         help='Override any kernel name specified in a notebook or request (KG_FORCE_KERNEL_NAME env var)')
+
     @default('force_kernel_name')
     def force_kernel_name_default(self):
         return os.getenv(self.force_kernel_name_env, '')
@@ -220,17 +240,43 @@ class KernelGatewayApp(JupyterApp):
             another fully qualified module name (KG_API env var)
             """
     )
+
     @default('api')
     def api_default(self):
         return os.getenv(self.api_env, self.api_default_value)
+
     @observe('api')
     def api_changed(self, event):
         try:
-            new_module = self._load_api_module(event['new'])
+            self._load_api_module(event['new'])
         except ImportError:
             # re-raise with more sensible message to help the user
             raise ImportError('API module {} not found'.format(event['new']))
 
+    certfile_env = 'KG_CERTFILE'
+    certfile = Unicode(None, config=True, allow_none=True,
+                       help="""The full path to an SSL/TLS certificate file. (KG_CERTFILE env var)""")
+
+    @default('certfile')
+    def certfile_default(self):
+        return os.getenv(self.certfile_env)
+
+    keyfile_env = 'KG_KEYFILE'
+    keyfile = Unicode(None, config=True, allow_none=True,
+                      help="""The full path to a private key file for usage with SSL/TLS. (KG_KEYFILE env var)""")
+
+    @default('keyfile')
+    def keyfile_default(self):
+        return os.getenv(self.keyfile_env)
+
+    client_ca_env = 'KG_CLIENT_CA'
+    client_ca = Unicode(None, config=True, allow_none=True,
+                        help="""The full path to a certificate authority certificate for SSL/TLS client authentication. (KG_CLIENT_CA env var)""")
+
+    @default('client_ca')
+    def client_ca_default(self):
+        return os.getenv(self.client_ca_env)
+
     def _load_api_module(self, module_name):
         """Tries to import the given module name.
 
@@ -244,7 +290,6 @@ class KernelGatewayApp(JupyterApp):
         module
             Module with the given name loaded using importlib.import_module
         """
-        api_module = None
         # some compatibility allowances
         if module_name == 'jupyter-websocket':
             module_name = 'kernel_gateway.jupyter_websocket'
@@ -383,6 +428,7 @@ class KernelGatewayApp(JupyterApp):
             kg_max_age=self.max_age,
             kg_max_kernels=self.max_kernels,
             kg_api=self.api,
+            kg_personality=self.personality,
             # Also set the allow_origin setting used by notebook so that the
             # check_origin method used everywhere respects the value
             allow_origin=self.allow_origin
@@ -395,17 +441,44 @@ class KernelGatewayApp(JupyterApp):
             if kg_name not in self.web_app.settings:
                 self.web_app.settings[kg_name] = trait_value.get(obj=self.personality)
             else:
-                self.log.warning('The personality trait name, %s, conflicts with a kernel gateway trait.' , trait_name)
+                self.log.warning('The personality trait name, %s, conflicts with a kernel gateway trait.', trait_name)
+
+    def _build_ssl_options(self):
+        """Build a dictionary of SSL options for the tornado HTTP server.
+
+        Taken directly from jupyter/notebook code.
+        """
+        ssl_options = {}
+        if self.certfile:
+            ssl_options['certfile'] = self.certfile
+        if self.keyfile:
+            ssl_options['keyfile'] = self.keyfile
+        if self.client_ca:
+            ssl_options['ca_certs'] = self.client_ca
+        if not ssl_options:
+            # None indicates no SSL config
+            ssl_options = None
+        else:
+            # SSL may be missing, so only import it if it's to be used
+            import ssl
+            # Disable SSLv3 by default, since its use is discouraged.
+            ssl_options.setdefault('ssl_version', ssl.PROTOCOL_TLSv1)
+            if ssl_options.get('ca_certs', False):
+                ssl_options.setdefault('cert_reqs', ssl.CERT_REQUIRED)
+
+        return ssl_options
 
     def init_http_server(self):
         """Initializes a HTTP server for the Tornado web application on the
         configured interface and port.
-        
+
         Tries to find an open port if the one configured is not available using
         the same logic as the Jupyer Notebook server.
         """
-        self.http_server = httpserver.HTTPServer(self.web_app)
-        
+        ssl_options = self._build_ssl_options()
+        self.http_server = httpserver.HTTPServer(self.web_app,
+                                                 ssl_options=ssl_options)
+
         for port in random_ports(self.port, self.port_retries+1):
             try:
                 self.http_server.listen(port, self.ip)
@@ -429,8 +502,8 @@ class KernelGatewayApp(JupyterApp):
     def start(self):
         """Starts an IO loop for the application."""
         super(KernelGatewayApp, self).start()
-        self.log.info('The Jupyter Kernel Gateway is running at: http://{}:{}'.format(
-            self.ip, self.port
+        self.log.info('Jupyter Kernel Gateway at http{}://{}:{}'.format(
+            's' if self.keyfile else '', self.ip, self.port
         ))
 
         self.io_loop = ioloop.IOLoop.current()
@@ -439,6 +512,8 @@ class KernelGatewayApp(JupyterApp):
             self.io_loop.start()
         except KeyboardInterrupt:
             self.log.info("Interrupted...")
+        finally:
+            self.shutdown()
 
     def stop(self):
         """
@@ -453,4 +528,5 @@ class KernelGatewayApp(JupyterApp):
         """Stop all kernels in the pool."""
         self.personality.shutdown()
 
+
 launch_instance = KernelGatewayApp.launch_instance
